from pysat.solvers import Glucose3

from famapy.core.operations import FalseOptionalFeatures
from famapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3FalseOptionalFeatures(FalseOptionalFeatures):

    def __init__(self):
        self.false_optional_features = []

    def get_false_optional_features(self):
        return self.false_optional_features

    def get_result(self):
        return self.get_false_optional_features()

    def execute(self, model: PySATModel) -> 'Glucose3FalseOptionalFeatures':
        glucose_r = Glucose3()
        glucose_r_ctc = Glucose3()
        for clause in model.r_cnf:
            glucose_r.add_clause(clause)
            glucose_r_ctc.add_clause(clause)
        for clause in model.ctc_cnf:
            glucose_r_ctc.add_clause(clause)

        if glucose_r_ctc.solve():
            assumption = 1
            for feat in model.features:
                if (not glucose_r_ctc.solve(assumptions=[-feat]) and 
                        glucose_r.solve(assumptions=[-feat])):
                    if glucose_r.solve(assumptions=[assumption, -feat]):
                        self.false_optional_features.append(model.features.get(feat))
                    assumption = feat

        glucose_r.delete()
        glucose_r_ctc.delete()
        return self



import sys

from famapy.core.exceptions import ElementNotFound
from famapy.core.models import VariabilityModel
from famapy.core.transformations import ModelToModel
from famapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class FmToPysat(ModelToModel):
    @staticmethod
    def get_source_extension():
        return 'fm'

    @staticmethod
    def get_destination_extension():
        return 'pysat'

    def __init__(self, source_model: VariabilityModel):
        self.source_model = source_model
        self.counter = 1
        self.destination_model = PySATModel()
        self.r_cnf = self.destination_model.r_cnf
        self.ctc_cnf = self.destination_model.ctc_cnf

    def add_feature(self, feature):
        if feature.name not in self.destination_model.variables.keys():
            self.destination_model.variables[feature.name] = self.counter
            self.destination_model.features[self.counter] = feature.name
            self.counter += 1

    def add_root(self, feature):
        self.r_cnf.append([self.destination_model.variables.get(feature.name)])

    def add_relation(self, relation):  # noqa: MC0001
        if relation.is_mandatory():
            self.r_cnf.append([
                -1 * self.destination_model.variables.get(relation.parent.name),
                self.destination_model.variables.get(relation.children[0].name)])
            self.r_cnf.append([
                -1 * self.destination_model.variables.get(relation.children[0].name),
                self.destination_model.variables.get(relation.parent.name)])

        elif relation.is_optional():
            self.r_cnf.append([
                -1 * self.destination_model.variables.get(relation.children[0].name),
                self.destination_model.variables.get(relation.parent.name)])

        elif relation.is_or():  # this is a 1 to n relatinship with multiple childs
            # add the first cnf child1 or child2 or ... or childN or no parent)

            # first elem of the constraint
            alt_cnf = [-1 * self.destination_model.variables.get(relation.parent.name)]
            for child in relation.children:
                alt_cnf.append(self.destination_model.variables.get(child.name))
            self.r_cnf.append(alt_cnf)

            for child in relation.children:
                self.r_cnf.append([
                    -1 * self.destination_model.variables.get(child.name),
                    self.destination_model.variables.get(relation.parent.name)])

        elif relation.is_alternative():  # this is a 1 to 1 relatinship with multiple childs
            # add the first cnf child1 or child2 or ... or childN or no parent)

            # first elem of the constraint
            alt_cnf = [-1 * self.destination_model.variables.get(relation.parent.name)]
            for child in relation.children:
                alt_cnf.append(self.destination_model.variables.get(child.name))
            self.r_cnf.append(alt_cnf)

            for i in range(len(relation.children)):
                for j in range(i + 1, len(relation.children)):
                    if i != j:
                        self.r_cnf.append([
                            -1 * self.destination_model.variables.get(relation.children[i].name),
                            -1 * self.destination_model.variables.get(relation.children[j].name)
                        ])
                self.r_cnf.append([
                    -1 * self.destination_model.variables.get(relation.children[i].name),
                    self.destination_model.variables.get(relation.parent.name)
                ])

        else:  # This is a m to n relationship
            print(
                "Fatal error. N to M relationships are not yet supported in PySAT",
                file=sys.stderr
            )
            raise NotImplementedError

    def add_constraint(self, ctc):
        dest = self.destination_model.variables.get(
            ctc.ast.get_childs(ctc.ast.get_root())[1].get_name()
        )
        orig = self.destination_model.variables.get(
            ctc.ast.get_childs(ctc.ast.get_root())[0].get_name()
        )
        if dest is None or orig is None:
            print(self.source_model)
            raise ElementNotFound
        if ctc.ast.get_root().get_name() == 'requires':
            self.r_cnf.append([-1 * orig, dest])

        elif ctc.ast.get_root().get_name() == 'excludes':
            self.r_cnf.append([-1 * orig, -1 * dest])

    def transform(self):
        for feature in self.source_model.get_features():
            self.add_feature(feature)

        self.add_root(self.source_model.root)

        for relation in self.source_model.get_relations():
            self.add_relation(relation)

        for constraint in self.source_model.get_constraints():
            self.add_constraint(constraint)

        return self.destination_model



from enum import Enum, auto

from famapy.core.transformations import TextToModel

from famapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class LogicOperator(Enum):
    NOT = auto()
    AND = auto()
    OR = auto()


class CNFNotation(Enum):
    LOGICAL = {LogicOperator.NOT: '¬', LogicOperator.AND: '∧', LogicOperator.OR: '∨'}
    JAVA = {LogicOperator.NOT: '!', LogicOperator.AND: '&&', LogicOperator.OR: '||'}
    SHORT = {LogicOperator.NOT: '-', LogicOperator.AND: '&', LogicOperator.OR: '|'}
    TEXTUAL = {LogicOperator.NOT: 'not', LogicOperator.AND: 'and', LogicOperator.OR: 'or'}


def identify_notation(cnf_formula: str) -> CNFNotation:
    for notation in CNFNotation:
        for symbol in notation.value.values():
            symbol_pattern = ' ' + symbol + ' '
            if symbol_pattern in cnf_formula:
                return notation
    raise Exception("This should have happend. \
            The CNF file is corrupted or does not adhere to the format") 


class CNFReader(TextToModel):
    """
    Read a CNF formula as a string representing a feature model.

    The expected format is the generated by FeatureIDE when exporting the model as CNF (.txt).
    That generates a file with the CNF formula in four different notations:
        Logical Symbols:
            (A) ∧ (¬B ∨ C) ∧ ...
        Textual Symbols:
            (A) and (not B or C) and ...
        Java Symbols:
            (A) && (!B || C) && ...
        Short Symbols:
            (A) & (-B | C) & ...
    This class is able to read any of these notations, but only one notation,
    so the .txt file should be modified to include only 
    one of those notations by removing the others.
    """

    @staticmethod
    def get_source_extension() -> str:
        return 'txt'

    def __init__(self, path: str):
        self._path = path
        self.counter = 1
        self.destination_model = PySATModel()
        self.cnf = self.destination_model.cnf

    def transform(self) -> PySATModel:
        self._read_clauses()
        return self.destination_model

    def _add_feature(self, feature_name):
        if feature_name not in self.destination_model.variables.keys():
            self.destination_model.variables[feature_name] = self.counter
            self.destination_model.features[self.counter] = feature_name
            self.counter += 1

    def _read_cnf_formula(self) -> str:
        """It assumes the CNF formula is defined in one line in the file."""
        with open(self._path, encoding='utf-8') as file:
            cnf_formula = file.readline()
        return cnf_formula

    def _read_clauses(self):
        cnf_formula = self._read_cnf_formula()
        cnf_notation = identify_notation(cnf_formula)

        and_symbol_pattern = ' ' + cnf_notation.value[LogicOperator.AND] + ' '
        clauses = list(map(lambda c: c[1:len(c) - 1], cnf_formula.split(and_symbol_pattern)))  
        # Remove initial and final parenthesis

        # Remove final parenthesis of last clause (because of the possible end of line: '\n')
        if ')' in clauses[len(clauses) - 1]:
            clauses[len(clauses) - 1] = clauses[len(clauses) - 1][:-1]  

        for _c in clauses:
            tokens = _c.split(' ')
            tokens = list(filter(lambda t: t != cnf_notation.value[LogicOperator.OR], tokens))
            logic_not = False
            cnf_clause = []
            for feature in tokens:
                if feature == cnf_notation.value[LogicOperator.NOT]:
                    logic_not = True
                elif feature.startswith(cnf_notation.value[LogicOperator.NOT]):
                    feature = feature.replace(cnf_notation.value[LogicOperator.NOT], '', 1)
                    self._add_feature(feature)
                    cnf_clause.append(-1 * self.destination_model.variables[feature])
                else:
                    self._add_feature(feature)
                    if logic_not:
                        cnf_clause.append(-1 * self.destination_model.variables[feature])
                    else:
                        cnf_clause.append(self.destination_model.variables[feature])
                    logic_not = False
            self.destination_model.add_constraint(cnf_clause)

    def get_cnf_formula(self, cnf_output_syntax: CNFNotation = CNFNotation.JAVA) -> str:
        cnf_formula = self._read_cnf_formula()
        cnf_notation = identify_notation(cnf_formula)

        if cnf_output_syntax == cnf_notation:
            return cnf_formula
        # Translate AND operators
        symbol_pattern = ' ' + cnf_notation.value[LogicOperator.AND] + ' '
        new_symbol = ' ' + cnf_output_syntax.value[LogicOperator.AND] + ' '
        cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

        # Translate OR operators
        symbol_pattern = ' ' + cnf_notation.value[LogicOperator.OR] + ' '
        new_symbol = ' ' + cnf_output_syntax.value[LogicOperator.OR] + ' '
        cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

        # Translate NOT operators (this is more complex because 
        # the symbol may be part of a feature's name)
        if cnf_notation == CNFNotation.TEXTUAL:
            symbol_pattern = cnf_notation.value[LogicOperator.NOT] + ' '
            new_symbol = cnf_output_syntax.value[LogicOperator.NOT]
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
        elif cnf_output_syntax == CNFNotation.TEXTUAL:
            symbol_pattern = ' ' + cnf_notation.value[LogicOperator.NOT]
            new_symbol = ' ' + cnf_output_syntax.value[LogicOperator.NOT] + ' '
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

            symbol_pattern = '(' + cnf_notation.value[LogicOperator.NOT]
            new_symbol = '(' + cnf_output_syntax.value[LogicOperator.NOT] + ' '
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
        else:
            symbol_pattern = ' ' + cnf_notation.value[LogicOperator.NOT]
            new_symbol = ' ' + cnf_output_syntax.value[LogicOperator.NOT]
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

            symbol_pattern = '(' + cnf_notation.value[LogicOperator.NOT]
            new_symbol = '(' + cnf_output_syntax.value[LogicOperator.NOT]
            cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
        return cnf_formula


from pysat.solvers import Glucose3

from famapy.core.operations import ErrorDetection
from famapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3ErrorDetection(ErrorDetection):

    def __init__(self):
        self.errors_messages = []

    def get_errors_messages(self):
        return self.errors_messages

    def get_result(self):
        return self.get_errors_messages()

    def execute(self, model: PySATModel) -> 'Glucose3ErrorDetection': # noqa: MC0001
        glucose_r = Glucose3()
        glucose_r_ctc = Glucose3()
        for clause in model.r_cnf:
            glucose_r.add_clause(clause)
            glucose_r_ctc.add_clause(clause)
        for clause in model.ctc_cnf:
            glucose_r_ctc.add_clause(clause)

        if glucose_r_ctc.solve():  # pylint:disable=too-many-nested-blocks
            dead_features = []
            for feat in model.features:
                if not glucose_r_ctc.solve(assumptions=[feat]):
                    dead_features.append(model.features.get(feat))
            if dead_features:
                self.errors_messages.append('Dead features: ' + str(dead_features))

            false_optional_features = []
            assumption = 1
            for feat in model.features:
                if (not glucose_r_ctc.solve(assumptions=[-feat]) and 
                        glucose_r.solve(assumptions=[-feat])):
                    if glucose_r.solve(assumptions=[assumption, -feat]):
                        false_optional_features.append(model.features.get(feat))
                    assumption = feat
            if false_optional_features:
                self.errors_messages.append('False optional \
                    features: ' + str(false_optional_features))

            redundancies = []
            for feat in model.features:
                if not glucose_r.solve(assumptions=[-feat]):
                    for clause in model.ctc_cnf:
                        if clause[1] == feat:
                            if clause[0] < 0:
                                redundancies.append(
                                    model.features.get(-clause[0]) + 
                                    ' requires ' + 
                                    model.features.get(feat)
                                )
                        if clause[0] == feat:
                            if clause[1] < 0:
                                redundancies.append(
                                    model.features.get(feat) + 
                                    ' requires ' + 
                                    model.features.get(-clause[1])
                                )

            variables = []
            for clause in model.ctc_cnf:
                if clause[1] in variables:
                    if clause[1] > 0:
                        redundancies.append(
                            model.features.get(abs(clause[0])) + 
                            ' requires ' + 
                            model.features.get(abs(clause[1]))
                        )
                    else:
                        redundancies.append(
                            model.features.get(abs(clause[0])) + 
                            ' excludes ' + 
                            model.features.get(abs(clause[1]))
                        )
                variables.append(clause[1])

            if redundancies:
                self.errors_messages.append('Redundancies: ' + str(redundancies))
        else:
            self.errors_messages.append('The model is void, so have not any product')

        glucose_r.delete()
        glucose_r_ctc.delete()
        return self



from pysat.solvers import Glucose3

from famapy.core.operations import ErrorDiagnosis
from famapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3ErrorDiagnosis(ErrorDiagnosis):

    def __init__(self):
        self.diagnosis_messages = []

    def get_diagnosis_messages(self):
        return self.diagnosis_messages

    def get_result(self):
        return self.get_diagnosis_messages()

    def execute(self, model: PySATModel) -> 'Glucose3ErrorDiagnosis': # noqa: MC0001
        glucose_r = Glucose3()
        glucose_r_ctc = Glucose3()
        for clause in model.r_cnf:
            glucose_r.add_clause(clause)
            glucose_r_ctc.add_clause(clause)
        for clause in model.ctc_cnf:
            glucose_r_ctc.add_clause(clause)

        if glucose_r_ctc.solve():
            dead_features = []
            for feat in model.features:
                if not glucose_r_ctc.solve(assumptions=[feat]):
                    dead_features.append(feat)

            false_optional_features = []
            assumption = 1
            for feat in model.features:
                if (not glucose_r_ctc.solve(assumptions=[-feat]) and 
                        glucose_r.solve(assumptions=[-feat])):
                    if glucose_r.solve(assumptions=[assumption, -feat]):
                        false_optional_features.append(feat)
                    assumption = feat

            diagnosis = []
            for dead in dead_features:
                name = model.features.get(dead)
                for clause in model.ctc_cnf:
                    if clause[1] == -dead:
                        if clause[0] < 0:
                            diagnosis.append(
                                'For dead feature ' + 
                                name + ': ' + 
                                model.features.get(-clause[0]) + 
                                ' excludes ' + 
                                name
                            )
                        else:
                            diagnosis.append(
                                'For dead feature ' + 
                                name + 
                                ': ' + 
                                model.features.get(clause[0]) + 
                                ' requires ' + 
                                name
                            )
                    if clause[0] == -dead:

                        if clause[1] < 0:
                            diagnosis.append(
                                'For dead feature ' + 
                                name + 
                                ': ' + 
                                name + 
                                ' excludes ' + 
                                model.features.get(-clause[1])
                            )
                        else:
                            diagnosis.append(
                                'For dead feature ' + 
                                name + 
                                ': ' + 
                                name + 
                                ' requires ' + 
                                model.features.get(clause[1])
                            )

            for false in false_optional_features:
                name = model.features.get(false)
                for clause in model.ctc_cnf:
                    if clause[1] == false:
                        if clause[0] < 0:
                            diagnosis.append(
                                'For false optional feature ' + 
                                name + ': ' + 
                                model.features.get(-clause[0]) + 
                                ' requires ' + 
                                name
                            )
                        else:
                            diagnosis.append(
                                'For false optional feature ' + 
                                name + 
                                ': ' + 
                                model.features.get(clause[0]) + 
                                ' excludes ' + 
                                name
                            )
                    if clause[0] == false:
                        if clause[1] < 0:
                            diagnosis.append(
                                'For false optional feature ' + 
                                name + 
                                ': ' + 
                                name + 
                                ' requires ' + 
                                model.features.get(-clause[1])
                            )
                        else:
                            diagnosis.append(
                                'For false optional feature ' + 
                                name + 
                                ': ' + 
                                name + 
                                ' excludes ' + 
                                model.features.get(clause[1])
                            )

            for message in diagnosis:
                if message not in self.diagnosis_messages:
                    self.diagnosis_messages.append(message)

        else:
            for clause in model.ctc_cnf:
                if clause[0] < 0 and clause[1] < 0:
                    self.diagnosis_messages.append(
                        model.features.get(-clause[0]) + 
                        ' excludes ' + 
                        model.features.get(-clause[1])
                    )

        glucose_r.delete()
        glucose_r_ctc.delete()
        return self


from pysat.formula import CNF

from famapy.core.models import VariabilityModel


class PySATModel(VariabilityModel):

    @staticmethod
    def get_extension():
        return 'pysat'

    def __init__(self):
        self.r_cnf = CNF()
        self.ctc_cnf = CNF()
        self.variables = {}
        self.features = {}

    def add_constraint(self, constraint):
        self.ctc_cnf.append(constraint)

    def get_all_clauses(self):
        clauses = CNF()
        clauses.extend(self.r_cnf.clauses)
        clauses.extend(self.ctc_cnf.clauses)
        return clauses